;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;                                               ;;;
;;;  Subtraction & Typing experiment              ;;;
;;;                                               ;;;
;;;  Jelmer Borst (jpborst@ai.rug.nl)             ;;;
;;;  Department of Artificial Intelligence        ;;;
;;;  University of Groningen                      ;;;
;;;  The Netherlands                              ;;;
;;;                                               ;;;
;;;  20080211                                     ;;;
;;;                                               ;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  
;;;  
;;; TODO:
;;; 
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Parameter definitions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;GUI parameters
;;;;;;;;;;;;;;;;;;;

;;empty cursor for letting the mouse 'disappear' when break in typing
(defparameter *empty-cursor* 
  (capi:load-cursor (merge-pathnames 
                     (make-pathname :name "emptycursor" :type "tif")  
                     (current-pathname)))) ;use current pathname to start with, merge with diff name 



;;Experiment flow parameters
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defparameter *block* 0) ;current block in the experiment
(defparameter *trial* 0)
(defparameter *sub_step* 0)
(defparameter *typ_step* 0)
(defparameter *done-flag* nil) ;flag to set when a block is finished - will stop the experiment or the model
(defparameter *points* 100)

; Rest
;;;;;;;;;;;;;

(defparameter *model* nil) ;are we using and ACT-R model yes/no
(defparameter *virtual* nil) ;display window for model yes/no
(defparameter *exp* nil) ;the experiment object
(defparameter *realtime* nil)



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; File Handling & participant setup - .../Lisp snippets/file handlers.lisp
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defparameter *pp* "test") ;participant#
(defparameter *exp-part* "test") ;name of the experiment part for use in filename

(defparameter *output-pathname* ;pathname to outputfile, default "pp-test"
  (merge-pathnames 
   (make-pathname :name "pp-test" :directory '(:relative "data") :type "dat")  
   (current-pathname))) ;use current path to start with, put it in relative directory 'data'

(defparameter *output-stream* nil) ;stream to write to
(defparameter *output-buffer* nil) ;buffer for output
(defparameter *output-buffer-size* 0) ;length of output buffer list
(defparameter *overwrite-p* nil) ;overwrite file if exist yes/no

;;ask for pp#, if exists, ask for overwrite confirmation or for new #
(defun setup-pp ()
  (setf *pp* (loop
              (let ((pp (capi:prompt-for-integer "Please enter a participant number:")))
                (cond 
                  ((probe-file ;if file exist
                    (setf *output-pathname* 
                          (make-pathname :name (format nil "~a-pp~a" *exp-part* pp) :defaults *output-pathname*)))
                   (when (capi:confirm-yes-or-no "Participant # already in use, overwrite and proceed anyway?") ;ask what to do
                     (setf *overwrite-p* t)
                     (return pp)))
                  (t
                   (return pp))))))) 


;;open the file to write and start with pp#, date and "(setf *data* '(", for easy reading later
(defun initialize-outfile ()
  (setf *output-pathname* (make-pathname :name (format nil "~a-pp~a" *exp-part* *pp*) :defaults *output-pathname*))
  (do ((i 1 (incf i)))
      ((setf *output-stream* ;actually don't overwrite when file already exists, but append -#.
             (open 
              (ensure-directories-exist *output-pathname*) 
              :direction :output 
              :if-does-not-exist :create 
              :if-exists (if *overwrite-p* ':supersede nil))))
    (setf *output-pathname* (make-pathname :name (format nil "~a-pp~a-~a" *exp-part* *pp* i) :defaults *output-pathname*)))
  (format *output-stream* "; Participant number: ~a~%" *pp*)
  (let ((time-list (multiple-value-list (get-decoded-time))))
    (format *output-stream* "; ~2,'0d:~2,'0d, ~d/~d/~2,'0d (mm/dd/yy)~%~%" 
            (third time-list) 
            (second time-list)
            (fifth time-list)
            (fourth time-list)
            (sixth time-list)))
  (format *output-stream* "(setf *data* '("))



;;write to buffer, if size exceeds 10: flush buffer to file
(defun print-outfile (string &rest args)
  (let ((string (apply #'format (append (list nil (substitute #\Tab #\Space string)) args))))
    (push string *output-buffer*))
  (incf *output-buffer-size*)
  (if (> *output-buffer-size* 10) (flush-buffer)))

;;flush buffer to file
(defun flush-buffer ()
  (dolist (string (reverse *output-buffer*))
    (format *output-stream* "~&")
    (format *output-stream* string))
  (setf *output-buffer* nil)
  (setf *output-buffer-size* 0)
  (gc-if-needed)) ;garbage cleaning if needed

;;flush buffer & close file and end the file with "))"
(defun finalize-outfile ()
  (flush-buffer)
  (format *output-stream* "))")
  (format *output-stream* "~%;points: ~a" (points *exp*))
  (close *output-stream*))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Supporting functions - switch on what's necessary
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun kill-toolbar ()
 (let ((interfaces (capi:collect-interfaces 'LISPWORKS-TOOLS::LISPWORKS-TOOLBAR-WINDOW))) ;destroy the lispworks toolbar!
    (if interfaces (mapcar #'(lambda (interface) 
                               (capi:apply-in-pane-process interface 'capi:destroy interface))
                           interfaces))))


(defun randomize (lst)		;list randomizer
  (let ((lst (copy-list lst)))
    (dotimes (i (length lst))
      (let ((x (nth i lst))
            (ni (+ i (random (- (length lst) i)))))
        (setf (nth i lst) (nth ni lst))
        (setf (nth ni lst) x)))
    lst))

;;returns time to use, real-time in case of human pp, act-r time in case of model
(defmacro my-get-time ()  `(if *model* (* (mp-time) 1000) (get-internal-real-time)))

(defmacro my/ (x y) `(if (zerop ,y) 0 (/ ,x ,y)))	;my div, 0 if div by 0
(defmacro sqr (x) `(let ((v ,x)) (* v v)))		;square

(defmacro minsigned (x m) `(let ((xv ,x) (mv ,m)) (if (>= xv 0) (min xv mv) (max xv (- mv)))))		;min op

(defun scramble (sequence) ;scrambles any sequence, including strings: "jelmer" -> "eljrme"
  (setf sequence (copy-seq sequence))
  (let ((length (length sequence)))
    (dotimes (i (1- length) sequence)
      (rotatef
       (elt sequence i)
       (elt sequence (random (+ i (- length  i))))))))

;;prints any list, one item per line
(defun print-list (lst)
  (dolist (i lst) (print i)))

;;starts act-r environment
(defun se ()
  (start-environment))

;;calculates number of different characters for two strings, used for calculating points
(defun diff (string1 string2)
  (let ((diffscore 0))
    (dotimes (i (length string1))
      (if (not (equalp (char string1 i) (char string2 i))) (incf diffscore)))
    diffscore))

(defun edit-distance (word1 word2)
  (let ((d (make-array (list 32 32) :element-type 'fixnum)))
    (declare (type simple-string word1 word2)
             (optimize (speed 3)))
    (let ((length1 (length word1))
          (length2 (length word2)))
      (loop for i from 0 to length1
            do (setf (aref d i 0) i))
      (loop for j from 0 to length2
            do (setf (aref d 0 j) j))
      (loop for i from 1 to length1
            do (loop for j from 1 to length2
                     do (setf (aref d i j)
                              (min (1+ (aref d (1- i) j))
                                   (1+ (aref d i (1- j)))
                                   (+ (aref d (1- i) (1- j))
                                      (if (char= (aref word1 (1- i))
                                                 (aref word2 (1- j)))
                                          0
                                        1))))))
      (aref d length1 length2))))


;;generate easy stimuli for subtraction, no carrying/borrowing.
(defun get-easy-stims (nr)
  (let ((lst nil))
    (dotimes (i nr)
      (push 
       (let ((A "0")
             (B "0"))
         (do () ((and ;keep trying to make such numbers until all terms are 10 digits long
                  (equalp (length (write-to-string (parse-integer A))) 10)
                  (equalp (length (write-to-string (parse-integer B))) 10)
                  (equalp (length (write-to-string (- (parse-integer A) (parse-integer B)))) 10)))
           (setf A "")
           (setf B "")
           (dotimes (i 10)
             (let ((termA (random 10))
                   (termB (random 10)))
               (do () ((>= termA termB)) ;choose random numbers, make sure B <= A
                 (setf termA (random 10))
                 (setf termB (random 10)))
               (setf A (format nil "~a~a" termA A))
               (setf B (format nil "~a~a" termB B)))))
         (list (parse-integer A) (parse-integer B)))
       lst))
    lst))

;;calculate how many times a carry is necessary in a subtraction
(defun calculate-carries (termA termB)
  (let ((A (reverse (format nil "~a" termA)))
        (B (reverse (format nil "~a" termB)))
        (prev1 "")
        (prev2 "")
        (carries 0))
    (loop 
     for dig1 across A
     for dig2 across B
     do
     (setf dig1 (parse-integer (format nil "~a" dig1)))
     (setf dig2 (parse-integer (format nil "~a" dig2)))
     (if (< dig1 dig2) (incf carries)) ;if dig1 < dig2, add carry
     (when (and
            (equalp dig1 dig2) ;if both digs are the equal
            (> (length prev1) 0) ;there already a previous digit
            (< (- (parse-integer prev1) (parse-integer prev2)) 0)) ;and the previous subtraction needs to carry
       (incf carries)) ;add a carry     
     (if (and (> (length prev1) 0) ;in case dig1 == 0 and carry is necessary, increase carries by 10, as we don't want that
              (equalp dig1 0) 
              (< (- (parse-integer prev1) (parse-integer prev2)) 0))
         (incf carries 10)) ;geen doorlopende carries!!

     (setf prev1 (format nil "~a~a" dig1 prev1))
     (setf prev2 (format nil "~a~a" dig2 prev2)))
    carries))


;;generate hard stimuli for subtraction, ie 6 carries in a 10 digit subtraction
(defun get-hard-stims (nr)
  (let ((lst nil))
    (dotimes (i nr)
      (push 
       (let ((A "0")
             (B "0"))
         (do () ((and
                  (equalp (length (write-to-string (parse-integer A))) 10)
                  (equalp (length (write-to-string (parse-integer B))) 10)
                  (equalp (length (write-to-string (- (parse-integer A) (parse-integer B)))) 10)
                  (> (- (parse-integer A) (parse-integer B)) 0)
                  (equalp (calculate-carries (parse-integer A) (parse-integer B)) 6)))
           (setf A "")
           (setf B "")
          ; (print (incf cnt))
           (dotimes (i 10)
             (let ((termA (random 10))
                   (termB (random 10)))
               (do () ((or (and (> (random 3) 1) (<= termA termB)) (>= termA termB)))
                 (setf termA (random 10))
                 (setf termB (random 10)))
               (setf A (format nil "~a~a" termA A))
               (setf B (format nil "~a~a" termB B)))))
         (list (parse-integer A) (parse-integer B)))
       lst))
    lst))


        



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Stimuli
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;stimuli for hard typing (CELEX)
(defparameter *typing-stimuli-hard*
  (randomize '("ziekenhuis"
               "veranderen"
               "behandelen"
               "belangrijk" 
               "verdwijnen" 
               "veranderen"
               "antwoorden" 
               "plotseling"
               "verwachten" 
               "nauwelijks" 
               "aanleiding"
               "beschouwen"
               "schreeuwen"
               "fluisteren"
               "historisch"
               "glimlachen" 
               "betrekking"
               "wetenschap" 
               "vervolgens"
               "verhouding"
               "economisch"
               "informatie"
               "activiteit"
               "vertrouwen"
              )))

;;stimuli for easy typing, scrambled hard stimuli without repeating same character
(defparameter *typing-stimuli-easy*
  (randomize '("ienzeiukhs"
               "naendvrere"
               "deaneblehn"
               "njgaleibrk" 
               "erniwjedvn"
               "rvnerdenea"
               "oanednowrt"
               "gelpilotns"
               "rhctwnveae"
               "enlskwaiju"
               "nlianigdae"
               "beshewuocn"
               "serwcehuen"
               "sernultief" 
               "osishirhct" 
               "hmlgacienl"
               "binkerkget"
               "wtsnhaecep"
               "evrelgsnov"
               "drnhivegou"
               "ocimonhcse"
               "nteomriaif"
               "taitetivic"
               "eoervuwnrt")))

;;easy subtraction stimuli
(defparameter *subtraction-stimuli-easy*
  (get-easy-stims 24))

;;hard subtraction stimuli
(defparameter *subtraction-stimuli-hard*
  (get-hard-stims 24))

;; list of difficulty levels for experiment
(defparameter *difficulty-levels* '())

;; 3 times (nr of blocks) a series of all conditions, 4 times per condition. That is, 3*3 = 9 times a condition, is 36 trials in total. 

(do ()
    ((= (length *difficulty-levels*) 36))
  (setf *difficulty-levels* nil)
  (let* ((lst '(1 2 3 4))
         (dif-list (randomize lst))
         (blocks 3) ;nr of blocks
         (out-lst nil))
    (dotimes (i (- blocks 1))
      (do ((new-lst (randomize lst) (randomize lst) )) 
          ((equalp (first new-lst) (nth (- (length dif-list) 1) dif-list)))
        (setf out-lst new-lst))
      (setf dif-list (append dif-list out-lst)))
    (dolist (i dif-list)
      (case i
        (1 ;;change lists to determine length of condition
           (setf *difficulty-levels* (append *difficulty-levels* '((easy easy) (easy easy) (easy easy))))) ; (easy easy)))))
        (2
         (setf *difficulty-levels* (append *difficulty-levels* '((easy hard) (easy hard) (easy hard))))) ; (easy hard)))))
        (3
         (setf *difficulty-levels* (append *difficulty-levels* '((hard easy) (hard easy) (hard easy))))) ; (hard easy)))))
        (4
         (setf *difficulty-levels* (append *difficulty-levels* '((hard hard) (hard hard) (hard hard))))))))) ; (hard hard)))))))))


;;param to hold difficulty levels for current block
;;experiment stops when this list is empty
(defparameter *current-levels* nil) 
                                     

;;;append stimuli for practice at front
(setf *subtraction-stimuli-easy* 
      (append (get-easy-stims 4)
              *subtraction-stimuli-easy*))
(setf *subtraction-stimuli-hard* 
      (append (get-hard-stims 4)
              *subtraction-stimuli-hard*))

(setf *typing-stimuli-easy* 
      (append 
       '("rjulitanuk" "poasgtnesi" "alkregvjia" "rebnecsmeh" "reitutlaru")
       *typing-stimuli-easy*))
(setf *typing-stimuli-hard* 
      (append 
       '("natuurlijk" "toepassing" "gevaarlijk" "beschermen" "literatuur") 
       *typing-stimuli-hard*))


;;;
;;;  Simulation Items // ACT-R
;;; 

;;upper item for all simulation items
(defclass st-window-item ()
  ((x :accessor x :initform 0 :initarg :x)
   (y :accessor y :initform 0 :initarg :y)
   (color :accessor color :initform 0 :initarg :color)
   (height :accessor height :initform 0 :initarg :height)
   (width :accessor width :initform 0 :initarg :width)
   (kind :accessor kind :initarg :kind)))

;text-feature, chunks will have the following slots, to be filled later
(defclass text-feature (st-window-item)
  ((text :accessor text :initform "" :initarg :text))
  (:default-initargs
      :kind 'text
      :height 5
      :width 5
      :color 'black))

;etc.
(defclass button-feature (st-window-item)
  ((label :accessor label :initform "" :initarg :label)
   (enabled :accessor enabled :initform t :initarg :enabled))
  (:default-initargs
      :kind 'button
      :height 5
      :width 20
      :color 'black))

(defparameter *typing-on* nil)
(defparameter *subtraction-on* nil)

;;;lists of features to be added to the visicon when necessary
(defparameter *delete-button* nil)
(defparameter *alphabet-buttons* nil)
(defparameter *button-features* nil)
(defparameter *typing-input-feature* nil)
(defparameter *subtraction-A-features* nil)
(defparameter *subtraction-B-features* nil)
(defparameter *subtraction-C-features* nil)


;;simulation for ACT-R
(defun create-features ()
  (setf *delete-button*
        (make-instance 'button-feature :x 899 :y 308 :label "Delete" :width 35))
  (setf *alphabet-buttons*
        (mapcar 
         #'(lambda (label x y) (make-instance 'button-feature :x x :y y :label label)) 
         (list "A" "B" "C" "D" "E" "F" "G" ;labels
               "H" "I" "J" "K" "L" "M" "N" 
               "O" "P" "Q" "R" "S" "T" "U" 
               "V" "W" "X" "Y" "Z")
         (mapcar #'(lambda (x) (+ 679 x))
                 (list 0 40 80 120 160 200 240 ;x-coords
                       0 40 80 120 160 200 240
                       0 40 80 120 160 200 240
                       0 40 80 120 160))
         (mapcar #'(lambda (x) (+ 328 x))
                 (list 5 5 5 5 5 5 5 ;y-coords
                       33 33 33 33 33 33 33 
                       61 61 61 61 61 61 61
                       89 89 89 89 89))))
  (setf *button-features* (append (list *delete-button*) *alphabet-buttons*))
  (setf *typing-input-feature*
        (make-instance 'text-feature :x 689 :y 308 :text "" :width 100))
  (setf *subtraction-A-features*
        (mapcar #'(lambda (x) (make-instance 'text-feature :x x :y 295 :text "0")) 
                (mapcar #'(lambda (x) (+ 25 x))
                        (list 0 25 50 75 100 125 150 175 200 225))))
  (setf *subtraction-B-features*
        (mapcar #'(lambda (x) (make-instance 'text-feature :x x :y 335 :text "0")) 
                (mapcar #'(lambda (x) (+ 25 x))
                        (list 0 25 50 75 100 125 150 175 200 225))))
  (setf *subtraction-C-features*
        (mapcar #'(lambda (x) (make-instance 'text-feature :x x :y 375 :text "0")) 
                (mapcar #'(lambda (x) (+ 25 x))
                        (list 0 25 50 75 100 125 150 175 200 225)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Interface - .../Lisp snippets/interface.lisp
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;
;;;  Simulation Window
;;;

(defparameter *expwin* nil) ;window object containing everything

;;push button panel for typing input
(defclass alphabet-panel (capi:push-button-panel) ;isa push button panel
  ()
  (:default-initargs
   :items (mapcar #'(lambda (text loc-x loc-y) ;create real buttons for exact positioning
                      (make-instance 'capi:push-button
                                     :text text 
                                     :visible-min-width 48 ;this doesn't seem to have much influence
                                     ;:visible-max-width 2
                                     :max-width t
                                     :max-height t
                                     :data text
                                     :x loc-x
                                     :y loc-y))
                  (list "A" "B" "C" "D" "E" "F" "G" ;labels
                        "H" "I" "J" "K" "L" "M" "N" 
                        "O" "P" "Q" "R" "S" "T" "U" 
                        "V" "W" "X" "Y" "Z")
                  (list 0 40 80 120 160 200 240 ;x-coords
                        0 40 80 120 160 200 240
                        0 40 80 120 160 200 240
                        0 40 80 120 160)
                  (list 5 5 5 5 5 5 5 ;y-coords
                        33 33 33 33 33 33 33 
                        61 61 61 61 61 61 61
                        89 89 89 89 89))
   :layout-class 'capi:pinboard-layout ;layout within push button panel
   :selection-callback 'process-input-button ;call input-button function when pressed
   :print-function 'string-capitalize
   :layout-args '(:visible-min-width 290 :visible-min-height 120))) ;args for panel

;; define the complete interface:
(capi:define-interface st-window ()
  ((subviews :accessor subs :initform nil) ;will hold object for ACT-R
   (mouse-pos :accessor mouse-pos :initform (vector 0 0))) ;mouse-pos used for act-r
  (:panes 
   (empty-pane-typing capi:output-pane ;empty pane for when typing is hidden
                   :visible-min-width 300 
                   :visible-min-height 160 
                   :accessor empty-pane-typing)
   (typing-border capi:rectangle ;border around typing-panel
               :accessor typing-border
               :x 683
               :y 299
               :width 302
               :height 162)
   (input-screen-typing capi:text-input-pane ;typing text box - where the stimuli appear
                        :accessor input-screen-typing
                        :text ""
                        :enabled :read-only
                        :visible-min-width 185
                        :visible-min-height 20
                        :x 10
                        :y 3
                        :font (gp:make-font-description 
                               :family "times" 
                               :size 18 
                               :weight :medium                         
                               :slant :roman)
                        :enabled :read-only)
   (alphabet alphabet-panel ;see class def above
             :accessor alphabet
             :x 5
             :y 38)
   (delete-button capi:push-button ;delete button, backspace in typing panel
                  :accessor delete-button
                  :text "Delete"
                  :data "Delete"
                  :visible-min-width 80
                  :foreground :red
                  :callback 'process-ok-delete-buttons
                  :x 215
                  :y 8)
   (subtraction-border capi:rectangle ;border around subtraction-panel
               :accessor subtraction-border
               :x 39
               :y 299
               :width 302
               :height 162)
   (subtraction-line capi:line-pinboard-object ;line for subtraction
                     :accessor subtraction-line
                     :start-x 0
                     :start-y 95
                     :end-x 265
                     :end-y 95
                     :graphics-args '(:thickness 3) )
   (subtraction-minus capi:line-pinboard-object ;minus for subtraction
                     :accessor subtraction-minus
                     :start-x 275
                     :start-y 95
                     :end-x 285
                     :end-y 95
                     :graphics-args '(:thickness 3) )
   (empty-pane-subtraction capi:output-pane ;empty pane for when subtraction is hidden
                   :visible-min-width 280 
                   :visible-min-height 140 
                   :accessor empty-pane-subtraction)
   (input-screen-subtraction capi:display-pane ;subtraction text box - where the response appears
                        :accessor input-screen-subtraction
                        :text " 0 0 0 0 0 0 0 0 0 0"
                        :enabled :read-only
                        :visible-min-width 262
                        :visible-min-height 25
                        :font (gp:make-font-description 
                               :family "times" ;"andale mono" 
                               :size 22
                               :PITCH :FIXED ;;!! to make sure spacing is the always the same width, also see termA & B
                               :weight :medium                         
                               :slant :roman)
                        :x 4
                        :y 105
                        :visible-border :outline)
   (subtraction-input-border capi:rectangle ;border around subtraction-panel
               :accessor subtraction-input-border
               :x 2
               :y 102
               :width 265
               :height 35)
   (subtraction-termA capi:display-pane ;first term of subtraction
                      :accessor subtraction-termA
                      :text "1 2 3 4 5 6 7 8 9 0"
                      :font (gp:make-font-description 
                             :family "times" 
                             :size 22
                             :PITCH :FIXED
                             :weight :medium                         
                             :slant :roman)
                      :x 4
                      :y 15)
   (subtraction-termB capi:display-pane ;second term of subtraction
                      :accessor subtraction-termB
                      :text "1 2 3 4 5 6 7 8 9 0"
                      :font (gp:make-font-description 
                             :family "times" 
                             :size 22
                             :pitch :fixed
                             :weight :medium                         
                             :slant :roman)
                      :x 4
                      :y 55)
   (subtraction-correct capi:display-pane ;text panel for when subtraction is done correctly
                     :accessor subtraction-correct
                     :text "Goed!"
                     :font (gp:make-font-description 
                            :family "times" 
                            :size 30
                            :weight :medium                         
                            :slant :roman)
                     :foreground :green
                     :x 100
                     :y 30)
   (subtraction-wrong capi:display-pane ;text panel for when subtraction is done wrong
                     :accessor subtraction-wrong
                     :text "Fout."
                     :font (gp:make-font-description 
                            :family "times" 
                            :size 30
                            :weight :medium                         
                            :slant :roman)
                     :foreground :red
                     :x 100
                     :y 25)
   (sub-plus-points capi:display-pane ;text panel for plus points to appear (if correct)
                    :accessor sub-plus-points
                    :text "+ 100 punten!"
                    :font (gp:make-font-description 
                           :family "times" 
                           :size 30
                           :weight :medium                         
                           :slant :roman)
                    :foreground :black
                    :x 50
                    :y 75)
   (sub-min-points capi:display-pane ;text panel for minus points to appear (if wrong)
                    :accessor sub-min-points
                    :text "- 50 punten"
                    :font (gp:make-font-description 
                           :family "times" 
                           :size 30
                           :weight :medium                         
                           :slant :roman)
                    :foreground :black
                    :x 50
                    :y 75)
   (typing-correct capi:display-pane ;text panel for when typing is done correct
                :accessor typing-correct
                :text "Goed!"
                :font (gp:make-font-description 
                       :family "times" 
                            :size 30
                            :weight :medium                         
                            :slant :roman)
                :foreground :green
                :x 100
                :y 25)
   (typing-wrong capi:display-pane ;text panel for when typing is wrong
                 :accessor typing-wrong
                 :text "Fout."
                 :font (gp:make-font-description 
                        :family "times" 
                        :size 30
                        :weight :medium                         
                        :slant :roman)
                 :foreground :red
                 :x 100
                 :y 25)
   (typ-plus-points capi:display-pane ;plus points for typing
                    :accessor typ-plus-points
                    :text "+ 100 punten!"
                    :font (gp:make-font-description 
                           :family "times" 
                           :size 30
                           :weight :medium                         
                           :slant :roman)
                    :foreground :black
                    :x 50
                    :y 75)
   (typ-min-points capi:display-pane ;minus points for typing
                    :accessor typ-min-points
                    :text "- 50 punten"
                    :font (gp:make-font-description 
                           :family "times" 
                           :size 30
                           :weight :medium                         
                           :slant :roman)
                    :foreground :black
                    :x 50
                    :y 75)   

   (points capi:display-pane ;text panel for displaying the points
           :accessor points
           :text (write-to-string *points*)
           :title "Punten:"
           :title-position :left
           :font (gp:make-font-description 
                  :family "times" 
                  :size 25 
                  :weight :medium                         
                  :slant :roman)
           :title-font   (gp:make-font-description 
                          :family "times" 
                          :size 25
                          :weight :medium                         
                          :slant :roman)
           :foreground :black
           :x 450
           :y 100))
  (:layouts 
   (pinboard capi:pinboard-layout  ;complete window layout
                '(typing-border subtraction-layout typing-layout subtraction-border points)
                :input-model '(((:character) process-keypresses))
                :accessor pinboard)
   (subtraction-layout capi:switchable-layout ;subtraction-layout, can be empty, input, correct or wrong
               '(empty-pane-subtraction subtraction-input-layout subtraction-correct-layout subtraction-wrong-layout) 
               :accessor subtraction-layout
               :x 45
               :y 300)
   (subtraction-input-layout capi:pinboard-layout ;input layout for subtraction
                             '(input-screen-subtraction subtraction-termA subtraction-termB subtraction-line subtraction-minus subtraction-input-border)
                             :accessor subtraction-input-layout)
   (subtraction-correct-layout capi:pinboard-layout ;subtraction layout for when answer is correct
                            '(subtraction-correct sub-plus-points)
                            :accessor subtraction-correct-layout)
   (subtraction-wrong-layout capi:pinboard-layout ;subtraction layout for when answer is wrong
                            '(subtraction-wrong sub-min-points)
                            :accessor subtraction-wrong-layout)
   (typing-layout capi:switchable-layout ;typing layout, input, empty, correct & wrong layouts
               '(empty-pane-typing typing-input-layout typing-correct-layout typing-wrong-layout)
               :accessor typing-layout
               :x 684
               :y 300)
   (typing-correct-layout capi:pinboard-layout ;layout to show when word is correct
                       '(typing-correct typ-plus-points)
                       :accessor typing-correct-layout)
   (typing-wrong-layout capi:pinboard-layout ;layot for when it's wrong
                        '(typing-wrong typ-min-points)
                        :accessor typing-wrong-layout)
   (typing-input-layout capi:pinboard-layout ;input stuff on typing layout
                        '(input-screen-typing delete-button alphabet)
                        :accessor typing-input-layout))
  (:default-initargs :title "Experiment" ;titel
   :visible-min-width 1024
   :visible-min-height 768
   :x 0
   :y 0
  ))



;;change the text in the a pane text-input-pane
(defun set-input-text (pane text)
  (capi:apply-in-pane-process 
   pane #'(setf capi:text-input-pane-text) text pane))

;;change text in typing input pane
(defun set-typing-text (text)
  (when *model*
    (setf (text *typing-input-feature*) text)
    (proc-display))
  (set-input-text (input-screen-typing *expwin*) (string-upcase text)))

;;change text in subtraction input pane
(defun set-subtraction-text (text)
  (let ((t2 (format nil "~a" text)) ;convert whatever to string
        (t3 "                    ")) ;make a new empty 20 space string
    (dotimes (i (length t2))
      (setf t3 (concatenate 'string t3 (subseq t2 i (+ i 1)) " "))
      (if *model* 
          (setf (text (nth (- 9 i) *subtraction-C-features*)) (subseq t2 (- (length t2) 1 i) (- (length t2) i)))))
    (when *model* 
        (dotimes (i (- 10 (length t2)))
          (setf (text (nth i *subtraction-C-features*)) ""))
        (proc-display))

    (setf t3 (subseq t3 (- (length t3) 20) (- (length t3) 1))) ;shorten t3 to 20 characters
    (capi:apply-in-pane-process 
     (input-screen-subtraction *expwin*) #'(setf capi:display-pane-text) t3 (input-screen-subtraction *expwin*))))

;;set term A text (requires 10 digit input!!!)
(defun set-term-A (text)
   (let ((t2 (write-to-string text)) ;convert numbers to a string first
        (t3 ""))
    (dotimes (i (length t2)) ;again push spaces in between characters
      (setf t3 (concatenate 'string t3 (subseq t2 i (+ i 1)) " "))
      (if *model* (setf (text (nth i *subtraction-A-features*))  (subseq t2 i (+ i 1))))) 
    (if *model* (proc-display))
    (setf t3 (subseq t3 0 (- (length t3) 1)))
    (capi:apply-in-pane-process 
     (subtraction-termA *expwin*) #'(setf capi:display-pane-text) t3 (subtraction-termA *expwin*))))

;;same as term A for term B
(defun set-term-B (text)
  (let ((t2 (write-to-string text))
        (t3 ""))
    (dotimes (i (length t2))
      (setf t3 (concatenate 'string t3 (subseq t2 i (+ i 1)) " "))
      (if *model* (setf (text (nth i *subtraction-B-features*))  (subseq t2 i (+ i 1)))))
    (if *model* (proc-display))
    (setf t3 (subseq t3 0 (- (length t3) 1)))
    (capi:apply-in-pane-process 
     (subtraction-termB *expwin*) #'(setf capi:display-pane-text) t3 (subtraction-termB *expwin*))))

;;set the number of points, again, make sure it's text first
(defun set-points-text (text)
  (setf text (format nil "~a" text))
  ;(format t "~a~&~a~&" text (points *exp*))
  (capi:apply-in-pane-process 
   (points *expwin*) #'(setf capi:display-pane-text) text (points *expwin*)))

;; set the color of the points to red if val is t, otherwise to black.
(defun set-color-points (val)
  (capi:apply-in-pane-process
   (points *expwin*) #'(setf capi:simple-pane-foreground) (if val :red :black) (points *expwin*)))

;;set the text of the subtraction wrong points
(defun set-subtraction-points-wrong (text)
  (setf text (format nil "+ ~a punten." text))
  (capi:apply-in-pane-process
   (sub-min-points *expwin*) #'(setf capi:display-pane-text) text (sub-min-points *expwin*)))

;;set the text of subtraction wrong
(defun set-subtraction-wrong-text (text)
  (setf text (format nil "~a" text))
  (capi:apply-in-pane-process
   (subtraction-wrong *expwin*) #'(setf capi:display-pane-text) text (subtraction-wrong *expwin*)))

(defun set-subtraction-wrong-place (x)
  (capi:execute-with-interface 
   *expwin*
   #'(lambda () (setf (capi:pinboard-pane-position  (subtraction-wrong *expwin*)) 
                      (values x 25)))))

(defun set-typing-points-wrong (text)
  (setf text (format nil "+ ~a punten." text))
  (capi:apply-in-pane-process
   (typ-min-points *expwin*) #'(setf capi:display-pane-text) text (typ-min-points *expwin*)))

;;set the text of subtraction wrong
(defun set-typing-wrong-text (text)
  (setf text (format nil "~a" text))
  (capi:apply-in-pane-process
   (typing-wrong *expwin*) #'(setf capi:display-pane-text) text (typing-wrong *expwin*)))

(defun set-typing-wrong-place (x)
  (capi:execute-with-interface 
   *expwin*
   #'(lambda () (setf (capi:pinboard-pane-position (typing-wrong *expwin*)) 
                      (values x 25)))))
  
;;get current text
(defun get-input-text (pane)
   (capi:text-input-pane-text pane))

;;get current typing text
(defun get-typing-text ()
  (get-input-text (input-screen-typing *expwin*)))

;;get current subtraction text
(defun get-subtraction-text ()
  (get-input-text (input-screen-subtraction *expwin*)))

;;enable/disable typing: enable all buttons and text input pane, and let the mouse cursor disappear
(defun enable-typing (val)
  (if (null val) (setf (typing-stop-time *exp*) (my-get-time)))
  (setf (typing-enabled *exp*) val)
  (when *model*
    (dolist (i *button-features*)
      (setf (color i) (if val 'black 'grey)))
    (setf (color *typing-input-feature*) (if val 'black 'grey)))
  (when (equalp 
         (capi:switchable-layout-visible-child (typing-layout *expwin*)) 
         (typing-input-layout *expwin*))
    (capi:set-button-panel-enabled-items (alphabet *expwin*)
                                         :set val)
    (capi:apply-in-pane-process (delete-button *expwin*) #'(setf capi:button-enabled) val (delete-button *expwin*))
    (capi:apply-in-pane-process 
     (input-screen-typing *expwin*) #'(setf capi:text-input-pane-enabled) val (input-screen-typing *expwin*))
    (capi:apply-in-pane-process (pinboard *expwin*) ;load the empty cursor if disabling
                                (lambda () (setf (capi:simple-pane-cursor  (pinboard *expwin*)) 
                                                 (if val ':default *empty-cursor*))))))


;;enable/disable subtractoin: enable subtraction input pane
(defun enable-subtraction (val)
  (if (null val) (setf (subtraction-stop-time *exp*) (my-get-time)))
  (setf (subtraction-enabled *exp*) val)
  (when *model*
    (dolist (i *subtraction-C-features*)
      (setf (color i) (if val 'black 'grey))))
  (when (equalp 
         (capi:switchable-layout-visible-child (subtraction-layout *expwin*)) 
         (subtraction-input-layout *expwin*))
    (setf (capi:pinboard-object-graphics-args (subtraction-input-border *expwin*))
          (list :foreground (if val :black :grey))) 
    (capi:redraw-pinboard-object (subtraction-input-border *expwin*) t) 
    (capi:apply-in-pane-process
     (input-screen-subtraction *expwin*) #'(setf capi:simple-pane-foreground) (if val :black :grey) (input-screen-subtraction *expwin*))))



;;switch to different pane in typing layout, options:'input 'empty 'correct 'wrong
(defun switch-typing-layout (switch-to)
  (case switch-to
    ('input
     (capi:apply-in-pane-process 
      (typing-layout *expwin*) #'(setf capi:switchable-layout-visible-child) 
      (typing-input-layout *expwin*) (typing-layout *expwin*))
     (enable-typing t)
     (if *model* (setf *typing-on* t))) ;set delete button, push 123/abc buttons below
    ('empty
     (capi:apply-in-pane-process 
      (typing-layout *expwin*) #'(setf capi:switchable-layout-visible-child) 
      (empty-pane-typing *expwin*) (typing-layout *expwin*))
     (if *model* (setf *typing-on* nil)))
    ('correct
     (capi:apply-in-pane-process 
      (typing-layout *expwin*) #'(setf capi:switchable-layout-visible-child) 
      (typing-correct-layout *expwin*) (typing-layout *expwin*))
     (if *model* (setf *typing-on* nil)))
    ('wrong
     (capi:apply-in-pane-process 
      (typing-layout *expwin*) #'(setf capi:switchable-layout-visible-child) 
      (typing-wrong-layout *expwin*) (typing-layout *expwin*))
     (if *model* (setf *typing-on* nil))))
  (if *model* (proc-display)))

;;switch to different pane in subtraction layout, options:'input 'empty 'correct 'wrong
(defun switch-subtraction-layout (switch-to)
  (case switch-to
    ('input
     (capi:apply-in-pane-process 
      (subtraction-layout *expwin*) #'(setf capi:switchable-layout-visible-child) 
      (subtraction-input-layout *expwin*) (subtraction-layout *expwin*))
     (enable-subtraction t)
     (if *model* (setf *subtraction-on* t)))
    ('empty
     (capi:apply-in-pane-process 
      (subtraction-layout *expwin*) #'(setf capi:switchable-layout-visible-child) 
      (empty-pane-subtraction *expwin*) (subtraction-layout *expwin*))
     (if *model* (setf *subtraction-on* nil)))
    ('correct
     (capi:apply-in-pane-process 
      (subtraction-layout *expwin*) #'(setf capi:switchable-layout-visible-child) 
      (subtraction-correct-layout *expwin*) (subtraction-layout *expwin*))
     (if *model* (setf *subtraction-on* nil)))
    ('wrong
     (capi:apply-in-pane-process 
      (subtraction-layout *expwin*) #'(setf capi:switchable-layout-visible-child) 
      (subtraction-wrong-layout *expwin*) (subtraction-layout *expwin*))
     (if *model* (setf *subtraction-on* nil))))
  (if *model* (proc-display)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Experiment code
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;create window, display and return
(defun create-st-window () 
  (let ((interfaces (capi:collect-interfaces 'st-window))) ;destroy all old st-window interfaces
    (if interfaces (mapcar #'(lambda (interface) 
                               (capi:apply-in-pane-process interface 'capi:destroy interface))
                           interfaces)))

  (setf *expwin* (make-instance 'st-window)) ;create a new st-window object
  (if (null *virtual*) (capi:display *expwin*)) ;display the window
  (if *model* (create-features))
  *expwin*) 
      
 



  
;; Event handling

;;proces keyboard input
(defun process-keypresses (pane mouse-x mouse-y key)
  (declare (ignore mouse-x) ;declarations against warnings
           (ignore mouse-y)
           (ignore pane))
  (format t "~&Key ~a pressed!" key)
  (when (and 
         (subtraction-enabled *exp*) ;there is subtraction is enabled
         (subtraction-stimulus *exp*)) ; (enable-subtraction t))and there's a stimulus 
    
    (let ((time (my-get-time)))
      (case key
        ((list #\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9)	;if key = 0123456789
         (if (null (typing-enabled *exp*)) (enable-typing t))
         (setf (subtraction-response *exp*) (format nil "~a~a" key (subtraction-response *exp*))) ;increase sbutraction response
         (set-subtraction-text (subtraction-response *exp*)) ;change the text in the window
         (print-outfile "( ~a ~a 4 ~a ~a key ~a ~a)" ;and print to outfile
                        time
                        (- time (start-time *exp*)) 
                        (subtraction-level *exp*)
                        (typing-level *exp*)
                        key
                        *trial*)
         (setf *sub_step* (length (subtraction-response *exp*)))
         (if (and (not (equalp (typing-level *exp*) 'none)) 
                  (not (>= (length (subtraction-response *exp*)) 10)))
             (enable-subtraction nil)) ;disable subtraction
         (when (>= (length (subtraction-response *exp*)) 10) ;when the length of the respone is 10 or longer (shouldn't happen)
           (print-outfile "( ~a ~a 5 ~a ~a finish-subtraction ~a ~a)" ;print finished plus response
                 time
                 (- time (start-time *exp*)) 
                 (subtraction-level *exp*)
                 (typing-level *exp*)
                 (subtraction-response *exp*)
                 *trial*)
           ;check if response is correct
           (cond ;if correct:
            ((equalp (- (first (subtraction-stimulus *exp*)) (second (subtraction-stimulus *exp*))) (parse-integer (subtraction-response *exp*)))
             (switch-subtraction-layout 'correct) ;switch the layout to correct
             (print-outfile "( ~a ~a 8 ~a ~a subtraction-acc 1  ~a)" ;print it to the outfile
                            time 
                            (- time (start-time *exp*)) 
                            (subtraction-level *exp*)
                            (typing-level *exp*)
                            *trial*)
             (incf (points *exp*) 100)) ;incf the points
            (t ;if not correct
             (let ((points (- 100 (* (diff ;calculate the difference
                               (write-to-string (- (first (subtraction-stimulus *exp*)) (second (subtraction-stimulus *exp*))))
                               (subtraction-response *exp*)) 10))))
               (set-subtraction-points-wrong points) ;use that to print the wrong points
               (set-subtraction-wrong-text (format nil "~a ~a." (/ (- 100 points) 10) (if (equalp points 90) "fout" "fouten"))) ;set the text right
               (switch-subtraction-layout 'wrong) ;swith to the wrong layout
               (set-subtraction-wrong-place (if (equalp points 90) 100 90))
               (print-outfile "( ~a ~a 8 ~a ~a subtraction-acc 0 ~a)" ;and print it to the outfile
                              time
                              (- time (start-time *exp*)) 
                              (subtraction-level *exp*)
                              (typing-level *exp*)
                              *trial*)
               (incf (points *exp*) points)))) ;incf the number of points
           (set-points-text (points *exp*)) ;in any case, print the current number of points to the screen
           (setf (subtraction-stimulus *exp*) nil))))))) ;and set stimulus to nil, meaning subtraction is finished


;;process ok & delete buttons
(defun process-ok-delete-buttons (data button)
  (declare (ignore button))
  (format t "~&OK-delete-Button ~a clicked!" data)
  (let ((time (my-get-time))
        (outstring nil))
    (cond
     ((equalp data "Delete")	;if button delete
      (when (typing-enabled *exp*) ;and typing is enabled
        (if (null (subtraction-enabled *exp*)) (enable-subtraction t))
        (let* ((txt (typing-response *exp*)) ;txt = response up till now
               (len (length txt)))
          (when (> len 0)	; if txt>0, otherwise there's nothing to delete
            (decf (cur-char *exp*)) ;decf the chur-character of the the response
            (let* ((cc (cur-char *exp*))
                   (new-cc (+ cc 1)))
              (cond ((equalp (typing-level *exp*) 'easy)
                     (set-typing-text (if (> new-cc (length (typing-stimulus *exp*))) ;if easy set new stimulus
                                          "OK"
                                        (subseq (typing-stimulus *exp*) cc new-cc))))
                    ((equalp (typing-level *exp*) 'hard) ;if hard don't do anything - no feedback!
                     nil))
              (setf (typing-response *exp*) (subseq (typing-response *exp*) 0 (- (length (typing-response *exp*)) 1))))))
        (unless outstring (setf outstring data))
        (if (not (equalp (subtraction-level *exp*) 'none)) (enable-typing nil)) ;disable typing
        (print-outfile "( ~a ~a 2 ~a ~a press \"~a\" ~a)" ;print to outfile
                       time
                       (- time (start-time *exp*))
                       (subtraction-level *exp*)
                       (typing-level *exp*)
                       outstring
                       *trial*)
        (if *model* (proc-display))))
     ((equalp data "Ok") ;;if button OK  ;;SHOULDNT BE USED ANYMORE, NO OK BUTTON!
      (unless outstring (setf outstring data))
      (print-outfile "( ~a ~a 3 ~a ~a press \"~a\" ~a)" ;print some stuff to the outfile
                     time
                     (- time (start-time *exp*))
                     (subtraction-level *exp*)
                     (typing-level *exp*)
                     outstring
                     *trial*)
      (print-outfile "( ~a ~a 6 ~a ~a finish-typing \"~a\" ~a)"
                 time
                 (- time (start-time *exp*)) 
                 (subtraction-level *exp*)
                 (typing-level *exp*)
                 (typing-response *exp*)
                 *trial*)
      (set-typing-text "")
      (cond
       ((equalp (typing-stimulus *exp*) (typing-response *exp*)) ;if response was correct increase points etc
        (switch-typing-layout 'correct)
        (print-outfile "( ~a ~a 8 ~a ~a typing-acc 1 ~a)"
                       time
                       (- time (start-time *exp*)) 
                       (subtraction-level *exp*)
                       (typing-level *exp*)
                       *trial*)                       
        (incf (points *exp*) 100))
       (t                                            ;if response was wrong decrease points
        (switch-typing-layout 'wrong)
        (print-outfile "( ~a ~a 8 ~a ~a typing-acc 0 ~a)"
                       time
                       (- time (start-time *exp*)) 
                       (subtraction-level *exp*)
                       (typing-level *exp*)
                       *trial*)
        (decf (points *exp*) 50)))
      (set-points-text (points *exp*)) ;set the points on the screen
      (setf (typing-stimulus *exp*) nil))))) ;set typing stimulus to nil, in effect ending typing
  



;;process button clicks of input panel: data can be a...z
(defun process-input-button (data button)
  (let ((time (my-get-time))
        (outstring nil)
        (exp *exp*))
    (when (typing-enabled *exp*) ;only process stuff if typing is enabled
      (if (null (subtraction-enabled exp)) (enable-subtraction t))
      (cond
       ((equalp data "Ok") ;if data="ok" then goto ok handler - no OK button anymore, shouldn't happen
        (process-ok-delete-buttons data button))
       (t ;else, do this:
        (format t "~&Input-Button ~a clicked!" data)
        (let* ((cc (cur-char exp))
               (new-cc (+ cc 1)))
          (cond 
           ((equalp (typing-level exp) 'easy)
            (set-typing-text (if (>= new-cc (length (typing-stimulus exp)))
                                ""
                               (string-upcase (subseq (typing-stimulus exp) new-cc (+ 1 new-cc))))))
           ((equalp (typing-level exp) 'hard)		;nav-hard
            (if (equalp cc 0) 
                (set-typing-text ""))))
          (setf (typing-response exp) (format nil "~a~a" (typing-response exp) data))
          (setf *typ_step* (length (typing-response *exp*)))
          (setf outstring data)
          (incf (cur-char exp)))
        (unless outstring (setf outstring data))
        (if (and (not (equalp (subtraction-level *exp*) 'none))
                 (not (>= (length (typing-response *exp*)) 10)))
            (enable-typing nil))
        (print-outfile "( ~a ~a 2 ~a ~a press \"~a\" ~a)"
                       time
                       (- time (start-time *exp*)) 
                       (subtraction-level *exp*)
                       (typing-level *exp*)
                       outstring
                       *trial*)
        (when (>= (length (typing-response *exp*)) 10) ;if 10 characters were entered
          (print-outfile "( ~a ~a 6 ~a ~a finish-typing \"~a\" ~a)"
                         time
                         (- time (start-time *exp*)) 
                         (subtraction-level *exp*)
                         (typing-level *exp*)
                         (typing-response *exp*)
                         *trial*)
          (set-typing-text "")
          (cond
           ((equalp (typing-stimulus *exp*) (typing-response *exp*)) ;if response was correct increase points etc
            (switch-typing-layout 'correct)
            (print-outfile "( ~a ~a 8 ~a ~a typing-acc 1 ~a)"
                           time
                           (- time (start-time *exp*)) 
                           (subtraction-level *exp*)
                           (typing-level *exp*)
                           *trial*)                       
            (incf (points *exp*) 100))
           (t                                            ;if response was wrong decrease points
            (let ((points (- 100 (* (edit-distance ;calculate the difference
                                     (string-upcase (typing-stimulus *exp*))
                                     (string-upcase (typing-response *exp*))) 10))))
              (set-typing-points-wrong points) ;use that to print the wrong points
              (set-typing-wrong-text (format nil "~a ~a." (/ (- 100 points) 10) (if (equalp points 90) "fout" "fouten"))) ;set the text right
              (switch-typing-layout 'wrong) ;swith to the wrong layout
              (set-typing-wrong-place (if (equalp points 90) 100 90))
              (print-outfile "( ~a ~a 8 ~a ~a typing-acc 0 ~a)"
                             time
                             (- time (start-time *exp*)) 
                             (subtraction-level *exp*)
                             (typing-level *exp*)
                             *trial*)
              (incf (points *exp*) points))))
          (set-points-text (points *exp*)) ;set the points on the screen
          (setf (typing-stimulus *exp*) nil))))
     ))) ;set typing stimulus t)))))
  
  

    
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  Experiment
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defclass experiment ()		;define class
  ((window :accessor window)	;with slots: experiment window
   (start-time :accessor start-time) ;start time of experiment
   (end-time :accessor end-time) ;end-time of a round of stimuli
   (points-time :accessor points-time :initform 0) ;last time of points decrease
   (typing-level :accessor typing-level :initform "none") ;typing level (none, easy or hard)
   (subtraction-level :accessor subtraction-level :initform "none") ;subtraction level (none easy or hard)
   (points :accessor points :initform *points*) ;current amount of points
   (busy :accessor busy :initform nil) ;busy responding to stimuli or not
   (typing-stimulus :accessor typing-stimulus :initform nil) ;current typing stimulus
   (typing-response :accessor typing-response :initform "") ;current typing response
   (subtraction-stimulus :accessor subtraction-stimulus :initform nil) ;current subtraction stimulus
   (subtraction-response :accessor subtraction-response :initform "") ;current subtraction response
   (cur-char :accessor cur-char) ;cur character (position) of response
   (typing-delay :accessor typing-delay :initform 2000) ;typing break length
   (typing-enabled :accessor typing-enabled :initform t) ;typing enabled yes/nop
   (typing-stop-time :accessor typing-stop-time :initform 0) ;last typing disable time
   (subtraction-delay :accessor subtraction-delay :initform 2000) ;subtraction break time
   (subtraction-enabled :accessor subtraction-enabled :initform t) ;subtraction enabled yes/no
   (subtraction-stop-time :accessor subtraction-stop-time :initform 0))) ;last subtraction stop time

;; this function is constantly called during the experiment
(defun experiment-step (exp)
  (let ((time (my-get-time)))
    (when (and (null (busy *exp*)) ;if there's no stimulus, and the last one is 5 seconds ago
               (< (+ (end-time *exp*) 5000) time)) 
      (start-new-trial *exp*)) ;start a new trial
  
    (when (busy exp) ;when there is a stimulus
      (when (< (+ (points-time *exp*) 500) time) ;when the points have not been updated for a second
        (decf (points *exp*)) ;decrease the points by one
       ; (format t "hier?~%")
        (set-points-text (points *exp*)) ;change points text
        (setf (points-time *exp*) time)) ;change last update time
    
      (when (and 
             (null (typing-stimulus *exp*)) ;when there are no stimuli anymore
             (null (subtraction-stimulus *exp*)))
        (end-trial exp)))))               ;end the trial


;;start a new trial, display stimuli etc
(defun start-new-trial (exp)
  (setf (busy exp) t) ;set busy to t
  (incf *trial*)
  (set-color-points t) ;set the color of the points to red
 
  (setf (cur-char exp) 0) ;set the cur char to 0
  (setf (subtraction-response exp) "") ;set both responses to ""
  (setf (typing-response exp) "")

  ;pop list of the *current-levels* list, and use it to set the levels of the tasks
  (let ((levels (pop *current-levels*)))
    (setf (subtraction-level exp) (first levels))
    (setf (typing-level exp) (second levels)))

  (case (typing-level exp) ; depending on the level, set the appropriate stimulus
    ('none
     (setf (typing-stimulus exp) nil))
    ('easy
     (setf (typing-stimulus exp) (pop *typing-stimuli-easy*))
     (set-typing-text (subseq (typing-stimulus exp) 0 1)))
    ('hard
     (setf (typing-stimulus exp) (pop *typing-stimuli-hard*))
     (set-typing-text (typing-stimulus exp))))
    
  
  (case (subtraction-level exp) ; depending on the level, set the appropriate stimulus
    ('none
     (setf (subtraction-stimulus exp) nil))
    ('easy
     (setf (subtraction-stimulus exp) (pop *subtraction-stimuli-easy*)))
    ('hard
     (setf (subtraction-stimulus exp) (pop *subtraction-stimuli-hard*))))
 
  (let ((time (my-get-time)))
    (when (not (equalp (subtraction-level exp) 'none)) ;if there's subtraction, set the window element correctly
      (set-term-A (first (subtraction-stimulus exp)))
      (set-term-B (second (subtraction-stimulus exp)))
      (set-subtraction-text "")
      (switch-subtraction-layout 'input)
      (enable-subtraction t)
      (print-outfile "( ~a ~a 0 ~a ~a new-subtraction ~a ~a)"
                     time
                     (- time (start-time *exp*)) 
                     (subtraction-level *exp*)
                     (typing-level *exp*)
                     (subtraction-stimulus exp)
                     *trial*))
    
    (when (not (equalp (typing-level exp) 'none)) ;if there's typing, set the window elements for that
      (switch-typing-layout 'input)
     ; (enable-typing t)
    
      (print-outfile "( ~a ~a 1 ~a ~a new-typing \"~a\" ~a)"
                     time
                     (- time (start-time *exp*)) 
                     (subtraction-level *exp*)
                     (typing-level *exp*)
                     (typing-stimulus exp)
                     *trial*))))
    
;end a trial
(defun end-trial (exp)
  (setf (busy exp) nil) ;switch off busy
  (set-color-points nil) ;set color of points to black
  (setf (end-time exp) (my-get-time)) ;set end-time
  (when (null *current-levels*) ;if there are no more levels, set done flag to true
    (sleep 5)
    (setf *done-flag* t)))
 
;; creates an experiment 
(defun create-experiment ()			
  (let ((exp (make-instance 'experiment)))	;object instance 'exp'
    (setf (window exp)      (create-st-window)	;set exp params, with a new window
          (start-time exp)  (my-get-time)
          (end-time exp) (my-get-time))		;the current time
    (if *model* (create-features))     			                ;and the current time again
    exp)) ;return exp-object

;; initialize expeirment
(defmethod experiment-initialize ((exp experiment)) ;initialize everything
  (setf *exp* exp) ;set *exp* to exp
  (setf (points *exp*) *points*)
  (setf *expwin* (window exp)) ;set *expwin* to the window of the experiment
  (setf *done-flag* nil)) ;set done nill

(defmethod experiment-run ((exp experiment))	;run experiment exp
  (experiment-initialize exp)			;initialize first
  (if *model* (install-device (window exp)))	;if model is running, set device
  (initialize-outfile)				;ini outfile, set path to write to
  (if (null *model*)            ;if no model
      (do () (*done-flag*)	;as long as done-flag is nil, do experiment-step
        (experiment-step exp))
    (progn                      ;if model
      (schedule-periodic-event .1 (lambda () (experiment-step exp)) :priority :max ) ;schedule exp step every 10 ms
      (run-until-condition (lambda () *done-flag*) :real-time *realtime*))) ;run until done flag is t
  (finalize-outfile)) ;finalize outfile

;; end experiment : destroy window
(defmethod experiment-end ((exp experiment))
  (setf *points* (points *exp*))
  (capi:destroy (window exp)))

;; start exp
(defun e (&optional name)
  (setf *exp-part* name)
  (let* ((exp (create-experiment)) ;create exp
         (res (experiment-run exp))) ;run exp
    (experiment-end exp)		;close window and go back to listener
    res))

;;; Run-Experiment

;;runs experiment while block < 7
(defun ex ()
  (kill-toolbar)
  (do () ((>= *block* 7))
      (experiment))
  (capi:display-message "Dat was alles, bedankt voor het meedoen!"))

;;real experiment method (case because of possible eyetrackign)
(defun experiment ()
  (case *block*
    
    ;;Subtraction-easy practice
    (0
     (setf *exp-part* "psubtractionEasy")
     (setup-pp)

     ;set correct params
     (setf *current-levels*
           '(
             (easy none)
             (easy none)
             (easy none)))
  
     (capi:display-message "Je gaat nu de makkelijke variant van de aftreksommen oefenen. Het is de bedoeling dat je je linkerhand gebruikt om de cijfers in te typen! Klik op OK om te beginnen.")
     (e "psubtractionEasy"))
    
    (1
     (setf *exp-part* "psubtractionHard")
    ;set correct params
     (setf *current-levels*
           '(
             (hard none)
             (hard none)
             (hard none)))
  

     (capi:display-message "Je gaat nu de moeilijke variant van de aftreksommen oefenen. Klik op OK om te beginnen.")
     (e "psubtractionHard"))
    
    (2
     (setf *exp-part* "ptypingEasy")
      ;set correct params
     (setf *current-levels*
           '(
             (none easy)
             (none easy)
             (none easy)))
    

     (capi:display-message  "Je gaat nu de makkelijke variant van het typen oefenen, gebruik hiervoor de muis. Klik op OK om te beginnen.")
     (e "ptypingEasy"))
   
    (3
     (setf *exp-part* "ptypingHard")
     ;set correct params
     (setf *current-levels*
           '(
             (none hard)
             (none hard)
             (none hard)))
   

     (capi:display-message "Je gaat nu de moeilijke variant van het typen oefenen. Klik op OK om te beginnen.")
     (e "ptypingHard"))
   
    (4
     (setf *exp-part* "pBoth")
     ;set correct params
     (setf *current-levels*
           '(
             (easy easy)
             (hard easy)
             (easy hard)
             (hard hard)))
    

     (capi:display-message "Je gaat nu beide taken tegelijk oefenen. Je krijgt vier trials: de eerste keer zijn beide taken makkelijk, dan worden de aftreksommen moeilijk, daarna zijn die weer makkelijk maar wordt het typen moeilijk, en tenslotte zijn beide taken moeilijk. Klik op OK om te beginnen.")
     (e "pBoth"))
   
    (5
     (setf *points* 200)
     (setf *trial* 0)
     (setf *exp-part* "Real1")
     ;set correct params
     (setf *current-levels*
           (subseq *difficulty-levels* 0 (/ (length *difficulty-levels*) 2)))
          
     (capi:display-message "Het echte experiment begint nu. Je krijgt steeds drie trials in dezelfde conditie, daarna wissel je van conditie. Veel succes!")
     (e "real1"))
    (6
     (setf *exp-part* "real2")
     ;set correct params
     (setf *current-levels*
           (subseq *difficulty-levels* (/ (length *difficulty-levels*) 2)))
          
     (capi:display-message "Tijd voor pauze! Je kunt verder gaan als je er klaar voor bent.")
     (e "real2")))
  (incf *block*))

;########################################################################
;### ACT-R interfacing
;########################################################################

;;;
;;; To start we will define the general motor methods needed for the device.
;;; In this case we will be moving the mouse so set the position methods
;;; appropriately.



(defmethod device-move-cursor-to ((device st-window) loc)
  (setf (mouse-pos device) loc))

(defmethod get-mouse-coordinates ((device st-window))
  (mouse-pos device))

(defmethod device-handle-click ((device st-window))
  (let* ((x (aref (get-mouse-coordinates *expwin*) 0))
         (y (aref (get-mouse-coordinates *expwin*) 1))
         (button nil))
    (dolist (i *button-features*)
      (if (and (equalp x (x i)) (equalp y (y i))) (setf button i)))
    
    (process-input-button (label button) button)))

(defmethod device-handle-keypress ((device st-window) key)
  (process-keypresses "actr" 0 0 key))

(defmethod device-speak-string ((devices st-window) string)
  ;; ignore model's speech output
  nil)



;;; Now we deal with the visual processing

;;; Still ignoring the mouse cursor for visual processing.

(defmethod cursor-to-vis-loc ((device st-window))
  nil)

;;; This method gets called during proc-display to generate
;;; the features available to the model.  It must return a 
;;; list (or list of lists) of visual-location (or sub-type of 
;;; visual-location) chunk names.  

;;; Here, we're calling specific build-vis-locs-for that
;;; are defined on the specific subviews and then just
;;; collecting those into a list to return.

(defmethod build-vis-locs-for ((device st-window) vis-mod)
  (setf (subs device) nil)
  (when *typing-on* 
    (dolist (x *button-features*) (push x (subs device)))
    (if (> (length (text *typing-input-feature*)) 0) (push *typing-input-feature* (subs device))))
  (when *subtraction-on*
    (dolist (x *subtraction-A-features*) (push x (subs device)))
    (dolist (x *subtraction-B-features*) (push x (subs device)))
    (dolist (x *subtraction-C-features*) (if (not (equalp (text x) "")) (push x (subs device)))))

  (let ((feats nil))
    (dolist (x (subs device) feats)
      (push (build-vis-locs-for x vis-mod) feats))))


(defmethod build-vis-locs-for ((view text-feature) vis-mod)
  (let ((vl (car (define-chunks-fct `((isa visual-location
                                           color ,(color view)
                                           screen-x ,(x view)
                                           screen-y ,(y view)
                                           height ,(height view)
                                           width ,(width view)
                                           kind ,(kind view)
                                           value ,(kind view)))))))
    
    
    (setf (chunk-real-visual-value vl) (text view))
    (setf (chunk-visual-object vl) view)
    vl))

(defmethod build-vis-locs-for ((view button-feature) vis-mod)
  (let ((vl (car (define-chunks-fct `((isa visual-location
                                           color ,(color view)
                                           screen-x ,(x view)
                                           screen-y ,(y view)
                                           height ,(height view)
                                           width ,(width view)
                                           kind ,(kind view)
                                           value ,(kind view)
                                           ))))))
        
    (setf (chunk-real-visual-value vl) (label view))
    (setf (chunk-visual-object vl) view)
    vl))

;(defmethod vis-loc-to-obj ((view text-feature) vis-loc)
;  (let ((new-object (car (define-chunks-fct `((isa ,(chunk-slot-value-fct vis-loc 'kind)))))))
;    (fill-default-vis-obj-slots new-object vis-loc)
;        (set-chunk-slot-value-fct new-object 'x (x view))
;        (set-chunk-slot-value-fct new-object 'y (y view))
;        new-object))